'**************************************************
' FILE:         Encoding
' AUTHOR:       Paulo Santos
' CREATED:      2005.JAN.13
' COPYRIGHT:    Copyright 2005  Paulo Santos
'               All Rights Reserved.
' DESCRIPTION:
'       Encoding Function
'
' MODIFICATION HISTORY:
' 01    2005.JAN.13
'       Paulo Santos
'       Added a New Encoding Class
' 
' 02    2005.JAN.14
'       Paulo Santos
'       Added summarized description of 
'       each component.
'
' 03    2005.FEB.20
'       Paulo Santos
'       Added an enhanced GetBytes function
'
' 04    2005.FEB.21
'       Paulo Santos
'       Added an enhanced GetChars function
'***************************************************

Namespace System.Text

    ''' <summary>
    ''' Helps to convert a ANSI input into a ANSI string.
    ''' </summary>
    Public Class ANSIEncoding
        Inherits System.Text.Encoding

        Private __ThreadsGetBytes As Integer
        Private __ThreadsGetChars As Integer

        Private Structure stateObjectGet
            Dim charArray() As Char
            Dim byteArray() As Byte
            Dim Index As Integer
            Dim Count As Integer
        End Structure

        ''' <summary>
        ''' Initializes a new instance of the ANSI Encoding class.
        ''' </summary>
        <DebuggerStepThrough()> _
        Public Sub New()
            MyBase.New()
        End Sub

        ''' <summary>
        ''' Gets the human-readable description of the encoding.
        ''' </summary>
        ''' <returns>The human-readable description of the encoding.</returns>
        <DebuggerHidden()> _
        Public Overrides ReadOnly Property EncodingName() As String
            Get
                Return "ANSI"
            End Get
        End Property

        ''' <summary>
        ''' Calculates the number of bytes required to encode a specified character array.
        ''' The ANSI Encoding takes into consideration all Unicode characters above the
        ''' byte boundary. For instance the Unicode character code 350 need two bytes to
        ''' be encoded.
        ''' </summary>
        ''' <param name="chars">The character array to encode.</param>
        ''' <param name="index">The starting index of the character array to encode.</param>
        ''' <param name="count">The numbers of characters to encode.</param>
        ''' <returns></returns>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetByteCount(ByVal chars() As Char, ByVal index As Integer, ByVal count As Integer) As Integer
            Return chars.Length
        End Function

        ''' <summary>
        ''' (Overloaded)
        ''' Encodes a range of characters from a character array into a byte array.
        ''' Each char in the array is converted to a byte through the Microsoft.VisualBasic.Asc() function.
        ''' </summary>
        ''' <param name="chars">The character array to encode.</param>
        ''' <param name="charIndex">The starting index of the character array to encode.</param>
        ''' <param name="charCount">The number of characters to encode.</param>
        ''' <param name="bytes">The byte array where the resulting encoding is stored.</param>
        ''' <param name="byteIndex">The starting index of the resulting encoding in the byte array.</param>
        ''' <returns>The number of bytes stored in the byte array.</returns>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetBytes(ByVal chars() As Char, ByVal charIndex As Integer, ByVal charCount As Integer, ByVal bytes() As Byte, ByVal byteIndex As Integer) As Integer

            '*
            '* Check all the parameters
            '*
            If (bytes Is Nothing) Then
                Throw New ArgumentNullException("bytes", "Byte array cannot be null (Nothing in Visual Basic).")
            End If

            If (chars Is Nothing) Then
                Throw New ArgumentNullException("chars", "Character array cannot be null (Nothing in Visual Basic).")
            End If

            If (byteIndex < 0) Then
                Throw New ArgumentOutOfRangeException("byteIndex", byteIndex, "byteIndex cannot be negative")
            End If

            If (charIndex < 0) Then
                Throw New ArgumentOutOfRangeException("charIndex", charIndex, "charIndex cannot be negative")
            End If

            If (charCount < 0) Then
                Throw New ArgumentOutOfRangeException("charCount", charCount, "charCount cannot be negative")
            End If

            Dim intPos As Integer = byteIndex
            For i As Integer = 0 To (charCount - 1)
                If (intPos >= bytes.Length) Then
                    Exit For
                End If
                If ((charIndex + i) >= chars.Length) Then
                    Exit For
                End If
                bytes(intPos) = Asc(chars(charIndex + i))
                intPos += 1
            Next

            Return (intPos - byteIndex)

        End Function

        ''' <summary>
        ''' (Overloaded)
        ''' Encodes a Char array into a Byte Array.
        ''' Each char in the array is converted to a byte through the Microsoft.VisualBasic.Asc() function.
        ''' </summary>
        ''' <param name="chars">The Char array to be encoded as an array of bytes.</param>
        ''' <returns>The encoded byte array.</returns>
        Public Overloads Function GetBytes(ByVal chars() As Char) As Byte()

            Dim pool As System.Threading.ThreadPool
            Dim state As New stateObjectGet
            Dim byteArray(chars.Length - 1) As Byte

            state.charArray = chars
            state.byteArray = byteArray

            '*
            '* Starts several threads in order to speed up the conversion.
            '*
            For i As Integer = 0 To chars.Length Step 256
                With state
                    .Index = i
                    If (chars.Length > 256) Then
                        .Count = Min(256, (chars.Length - i))
                    Else
                        .Count = chars.Length
                    End If
                End With
                __ThreadsGetBytes += 1
                Do While Not pool.QueueUserWorkItem(AddressOf Me.doGetBytes, state)
                    System.Threading.Thread.Sleep(100)
                Loop
            Next

            '*
            '* Wait until all threads are finished
            '*
            Do While (__ThreadsGetBytes > 0)
                System.Threading.Thread.Sleep(100)
            Loop

            Return (byteArray)

        End Function

        ''' <summary>
        ''' (Overloaded)
        ''' Encodes a String into a Byte Array.
        ''' Each char in the array is converted to a byte through the Microsoft.VisualBasic.Asc() function.
        ''' </summary>
        ''' <param name="s">A System.String to be encoded as an array of bytes.</param>
        ''' <returns>The encoded byte array.</returns>
        Public Overloads Overrides Function GetBytes(ByVal s As String) As Byte()
            Return GetBytes(s.ToCharArray)
        End Function

        ''' <summary>
        ''' Calculates the number of characters produced by decoding an array of bytes.
        ''' </summary>
        ''' <param name="bytes">The byte array to decode.</param>
        ''' <param name="index">The starting index where decoding begins.</param>
        ''' <param name="count">The number of bytes to decode.</param>
        ''' <returns>The number of characters produced by decoding an array of bytes.</returns>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetCharCount(ByVal bytes() As Byte, ByVal index As Integer, ByVal count As Integer) As Integer
            Return count
        End Function

        ''' <summary>
        ''' Decodes a byte array into an array of characters.
        ''' </summary>
        ''' <param name="bytes">The byte array to decode.</param>
        ''' <param name="byteIndex">The starting index of the byte array to decode.</param>
        ''' <param name="byteCount">The number of bytes to decode.</param>
        ''' <param name="chars">The character array where the resulting decoding is stored.</param>
        ''' <param name="charIndex">The starting index of the resulting decoding in the character array.</param>
        ''' <returns>An array of chars that contains the decoded bytes.</returns>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetChars(ByVal bytes() As Byte, ByVal byteIndex As Integer, ByVal byteCount As Integer, ByVal chars() As Char, ByVal charIndex As Integer) As Integer
            '*
            '* Check all the parameters
            '*
            If (bytes Is Nothing) Then
                Throw New ArgumentNullException("bytes", "Byte array cannot be null (Nothing in Visual Basic).")
            End If

            If (chars Is Nothing) Then
                Throw New ArgumentNullException("chars", "Character array cannot be null (Nothing in Visual Basic).")
            End If

            If (byteIndex < 0) Then
                Throw New ArgumentOutOfRangeException("byteIndex", byteIndex, "byteIndex cannot be negative")
            End If

            If (byteCount < 0) Then
                Throw New ArgumentOutOfRangeException("byteCount", byteCount, "byteCount cannot be negative")
            End If

            If (charIndex < 0) Then
                Throw New ArgumentOutOfRangeException("charIndex", charIndex, "charIndex cannot be negative")
            End If

            Dim intPos As Integer = byteIndex
            For i As Integer = 0 To (byteCount - 1)
                If (intPos >= bytes.Length) Then
                    Exit For
                End If
                If ((charIndex + i) >= chars.Length) Then
                    Exit For
                End If
                chars(charIndex + i) = Chr(bytes(intPos))
                intPos += 1
            Next

            Return (intPos - byteIndex)
        End Function

        ''' <summary>
        ''' Decodes a byte array into an array of characters.
        ''' </summary>
        ''' <param name="bytes">The byte array to decode.</param>
        ''' <returns>An array of chars that contains the decoded bytes.</returns>
        Public Overloads Overrides Function GetChars(ByVal bytes() As Byte) As Char()

            Dim pool As System.Threading.ThreadPool
            Dim state As New stateObjectGet
            Dim charArray(bytes.Length - 1) As Char

            state.charArray = charArray
            state.byteArray = bytes

            '*
            '* Starts several threads in order to speed up the conversion.
            '*
            For i As Integer = 0 To bytes.Length Step 256
                With state
                    .Index = i
                    If (bytes.Length > 256) Then
                        .Count = Min(256, (bytes.Length - i))
                    Else
                        .Count = bytes.Length
                    End If
                End With
                __ThreadsGetChars += 1
                Do While Not pool.QueueUserWorkItem(AddressOf Me.doGetChars, state)
                    System.Threading.Thread.Sleep(100)
                Loop
            Next

            '*
            '* Wait until all threads are finished
            '*
            Do While (__ThreadsGetChars > 0)
                System.Threading.Thread.Sleep(100)
            Loop

            Return (charArray)

        End Function

        ''' <summary>
        ''' Returns the maximum number of bytes required to encode a given number of characters.
        ''' The ANSI Encoding assumes that charCount are in UTF8 characters, therefore 
        ''' returns the same number.
        ''' </summary>
        ''' <param name="charCount">The number of characters to encode.</param>
        ''' <returns>The maximum number of bytes required for encoding a given number of characters.</returns>
        <DebuggerStepThrough()> _
        Public Overrides Function GetMaxByteCount(ByVal charCount As Integer) As Integer
            Return charCount
        End Function

        ''' <summary>
        ''' Returns the maximum number of characters produced by decoding a given number of bytes.
        ''' The ANSI Encoding assumes that charCount are in UTF8 characters, therefore 
        ''' returns the same number.
        ''' </summary>
        ''' <param name="byteCount">The number of bytes to encode.</param>
        ''' <returns>The maximum number of characters produced by decoding a specified number of bytes.</returns>
        <DebuggerStepThrough()> _
        Public Overrides Function GetMaxCharCount(ByVal byteCount As Integer) As Integer
            Return byteCount
        End Function

        ''' <summary>
        ''' Decodes the specified byte array into a string.
        ''' </summary>
        ''' <param name="bytes">The byte array to decode.</param>
        ''' <returns>A string containing the decoded representation of the specified byte array.</returns>
        <DebuggerStepThrough()> _
        Public Overloads Shared Function GetString(ByVal bytes() As Byte) As String
            Return (New String((New ANSIEncoding).GetChars(bytes)))
        End Function

        ''' <summary>
        ''' (Private)
        ''' Returns the smaller of two numbers.
        ''' </summary>
        ''' <param name="a">The first number to compare.</param>
        ''' <param name="b">The second number to compare.</param>
        ''' <returns>Rhe smaller of two numbers.</returns>
        Private Function Min(ByVal a As Integer, ByVal b As Integer) As Integer
            If (a < b) Then
                Return a
            Else
                Return b
            End If
        End Function

        ''' <summary>
        ''' (Private)
        ''' This function is used by GetBytes to speed up the encoding.
        ''' </summary>
        ''' <param name="stateObject">An object that contains the data to process.</param>
        Private Sub doGetBytes(ByVal stateObject As Object)
            Dim x As stateObjectGet = CType(stateObject, stateObjectGet)
            Call GetBytes(x.charArray, x.Index, x.Count, x.byteArray, x.Index)
            __ThreadsGetBytes -= 1
        End Sub

        ''' <summary>
        ''' (Private)
        ''' This function is used by GetChars to speed up the decoding.
        ''' </summary>
        ''' <param name="stateObject">An object that contains the data to process.</param>
        Private Sub doGetChars(ByVal stateObject As Object)
            Dim x As stateObjectGet = CType(stateObject, stateObjectGet)
            Call GetChars(x.byteArray, x.Index, x.Count, x.charArray, x.Index)
            __ThreadsGetChars -= 1
        End Sub

    End Class

End Namespace
